class RapidataDataTypes:
    # deprecated use Literal["media", "text"] instead
    MEDIA = "media"  # any form of image, video or audio
    TEXT = "text"

    @classmethod
    def _possible_values(cls):
        return [cls.MEDIA, cls.TEXT]



import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from rapidata.rapidata_client.config.rapidata_config import rapidata_config


class SessionManager:
    _session = None

    @classmethod
    def get_session(
        cls,
    ) -> requests.Session:
        """Get a singleton requests session with retry logic.

        Returns:
            requests.Session: A singleton requests session with retry logic.
        """
        if cls._session is None:
            max_retries: int = rapidata_config.upload.maxRetries
            max_workers: int = rapidata_config.upload.maxWorkers
            cls._session = requests.Session()
            retries = Retry(
                total=max_retries,
                backoff_factor=1,
                status_forcelist=[500, 502, 503, 504],
                allowed_methods=["GET"],
                respect_retry_after_header=True,
            )

            adapter = HTTPAdapter(
                pool_connections=max_workers * 2,
                pool_maxsize=max_workers * 4,
                max_retries=retries,
            )
            cls._session.mount("http://", adapter)
            cls._session.mount("https://", adapter)

        return cls._session



from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.datapoints.assets import MediaAsset
from rapidata.api_client.models.create_demographic_rapid_model import (
    CreateDemographicRapidModel,
)
from rapidata.api_client.models.classify_payload import ClassifyPayload
from rapidata.rapidata_client.config import logger


class DemographicManager:
    def __init__(self, openapi_service: OpenAPIService):
        self._openapi_service = openapi_service
        logger.debug("DemographicManager initialized")

    def create_demographic_rapid(
        self, instruction: str, answer_options: list[str], datapoint: str, key: str
    ):

        media = MediaAsset(path=datapoint)
        model = CreateDemographicRapidModel(
            key=key,
            payload=ClassifyPayload(
                _t="ClassifyPayload",
                possibleCategories=answer_options,
                title=instruction,
            ),
        )

        self._openapi_service.rapid_api.rapid_demographic_post(
            model=model, file=[media.to_file()]
        )



from typing import Literal, Optional, Sequence
import random
import urllib.parse
import webbrowser

from colorama import Fore
from rapidata.api_client.models.ab_test_selection_a_inner import AbTestSelectionAInner
from rapidata.api_client.models.and_user_filter_model_filters_inner import (
    AndUserFilterModelFiltersInner,
)
from rapidata.api_client.models.create_order_model import CreateOrderModel
from rapidata.api_client.models.create_order_model_referee import (
    CreateOrderModelReferee,
)
from rapidata.api_client.models.create_order_model_workflow import (
    CreateOrderModelWorkflow,
)
from rapidata.api_client.models.sticky_state import StickyState

from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.rapidata_client.exceptions.failed_upload_exception import (
    FailedUploadException,
    _parse_failed_uploads,
)
from rapidata.rapidata_client.filter import RapidataFilter
from rapidata.rapidata_client.config import (
    logger,
    managed_print,
    rapidata_config,
    tracer,
)
from rapidata.rapidata_client.validation.validation_set_manager import (
    ValidationSetManager,
)
from rapidata.rapidata_client.order.dataset._rapidata_dataset import RapidataDataset
from rapidata.rapidata_client.order.rapidata_order import RapidataOrder
from rapidata.rapidata_client.referee import Referee
from rapidata.rapidata_client.referee._naive_referee import NaiveReferee
from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.rapidata_client.settings import RapidataSetting
from rapidata.rapidata_client.workflow import Workflow, FreeTextWorkflow
from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.api.rapidata_api_client import (
    suppress_rapidata_error_logging,
)


class RapidataOrderBuilder:
    """Builder object for creating Rapidata orders.

    Use the fluent interface to set the desired configuration. Add a workflow to the order using `.workflow()` and finally call `.create()` to create the order.

    Args:
        openapi_service (OpenAPIService): The OpenAPIService instance.
        name (str): The name of the order.
    """

    def __init__(
        self,
        name: str,
        openapi_service: OpenAPIService,
    ):
        self._name = name
        self.order_id: str | None = None
        self.__openapi_service = openapi_service
        self.__dataset: Optional[RapidataDataset] = None
        self.__workflow: Workflow | None = None
        self.__referee: Referee | None = None
        self.__validation_set_id: str | None = None
        self.__settings: Sequence[RapidataSetting] | None = None
        self.__user_filters: list[RapidataFilter] = []
        self.__selections: list[RapidataSelection] = []
        self.__priority: int | None = None
        self.__datapoints: list[Datapoint] = []
        self.__sticky_state_value: Literal["None", "Temporary", "Permanent"] | None = (
            None
        )
        self.__validation_set_manager: ValidationSetManager = ValidationSetManager(
            self.__openapi_service
        )

    def _to_model(self) -> CreateOrderModel:
        """
        Convert the builder configuration to a CreateOrderModel.

        Raises:
            ValueError: If no workflow is provided.

        Returns:
            CreateOrderModel: The model representing the order configuration.
        """
        if self.__workflow is None:
            raise ValueError("You must provide a workflow to create an order.")

        if self.__referee is None:
            managed_print("No referee provided, using default NaiveReferee.")
            self.__referee = NaiveReferee()

        return CreateOrderModel(
            _t="CreateOrderModel",
            orderName=self._name,
            workflow=CreateOrderModelWorkflow(self.__workflow._to_model()),
            userFilters=[
                AndUserFilterModelFiltersInner(user_filter._to_model())
                for user_filter in self.__user_filters
            ],
            referee=CreateOrderModelReferee(self.__referee._to_model()),
            validationSetId=self.__validation_set_id,
            featureFlags=(
                [setting._to_feature_flag() for setting in self.__settings]
                if self.__settings is not None
                else None
            ),
            selections=(
                [
                    AbTestSelectionAInner(selection._to_model())
                    for selection in self.__selections
                ]
                if self.__selections
                else None
            ),
            priority=self.__priority,
            stickyState=(
                StickyState(self.__sticky_state_value)
                if self.__sticky_state_value
                else None
            ),
        )

    def _set_validation_set_id(self) -> bool:
        """
        Get the validation set ID for the order.

        Returns:
            bool: True if a new validation set was created, False otherwise.
        """
        assert self.__workflow is not None
        if self.__validation_set_id:
            logger.debug(
                "Using specified validation set with ID: %s", self.__validation_set_id
            )
            return False

        try:
            with suppress_rapidata_error_logging():
                self.__validation_set_id = (
                    (
                        self.__openapi_service.validation_api.validation_set_recommended_get(
                            asset_type=[self.__datapoints[0].get_asset_type()],
                            modality=[self.__workflow.modality],
                            prompt_type=[
                                t.value for t in self.__datapoints[0].get_prompt_type()
                            ],
                        )
                    )
                    .validation_sets[0]
                    .id
                )
            logger.debug(
                "Using recommended validation set with ID: %s", self.__validation_set_id
            )
            return False
        except Exception as e:
            logger.debug("No recommended validation set found, error: %s", e)

        if (
            len(self.__datapoints)
            < rapidata_config.order.minOrderDatapointsForValidation
        ):
            logger.debug(
                "No recommended validation set found, dataset too small to create one."
            )
            return False

        logger.info("No recommended validation set found, creating new one.")

        managed_print()
        managed_print(f"No recommended validation set found, new one will be created.")
        validation_set = self.__validation_set_manager._create_order_validation_set(
            workflow=self.__workflow,
            order_name=self._name,
            datapoints=random.sample(
                self.__datapoints,
                min(
                    rapidata_config.order.autoValidationSetSize, len(self.__datapoints)
                ),
            ),
            settings=self.__settings,
        )

        logger.debug("New validation set created for order: %s", validation_set)
        self.__validation_set_id = validation_set.id
        return True

    def _create(self) -> RapidataOrder:
        """
        Create the Rapidata order by making the necessary API calls based on the builder's configuration.

        Raises:
            ValueError: If both media paths and texts are provided, or if neither is provided.
            AssertionError: If the workflow is a CompareWorkflow and media paths are not in pairs.

        Returns:
            RapidataOrder: The created RapidataOrder instance.
        """
        if rapidata_config.order.autoValidationSetCreation and not isinstance(
            self.__workflow, FreeTextWorkflow
        ):
            new_validation_set = self._set_validation_set_id()
        else:
            new_validation_set = False

        order_model = self._to_model()
        logger.debug("Creating order with model: %s", order_model)

        result = self.__openapi_service.order_api.order_post(
            create_order_model=order_model
        )

        self.order_id = str(result.order_id)
        logger.debug("Order created with ID: %s", self.order_id)

        if rapidata_config.order.autoValidationSetCreation and new_validation_set:
            required_amount = min(int(len(self.__datapoints) * 0.01) or 1, 10)
            managed_print()
            managed_print(
                Fore.YELLOW
                + f"A new validation set was created. Please annotate {required_amount} datapoint{('s' if required_amount != 1 else '')} so the order runs correctly."
                + Fore.RESET
            )
            link = f"https://app.{self.__openapi_service.environment}/validation-set/detail/{self.__validation_set_id}/annotate?orderId={self.order_id}&required={required_amount}"
            could_open_browser = webbrowser.open(link)
            if not could_open_browser:
                encoded_url = urllib.parse.quote(link, safe="%/:=&?~#+!$,;'@()*[]")
                managed_print(
                    Fore.RED
                    + f"Please open this URL in your browser: '{encoded_url}'"
                    + Fore.RESET
                )
            managed_print(
                "If you want to avoid the automatic validation set creation in the future, set `rapidata_config.order.autoValidationSetCreation = False`."
            )
            managed_print()

        self.__dataset = (
            RapidataDataset(result.dataset_id, self.__openapi_service)
            if result.dataset_id
            else None
        )
        if self.__dataset:
            logger.debug("Dataset created with ID: %s", self.__dataset.id)
        else:
            logger.warning("No dataset created for this order.")

        order = RapidataOrder(
            order_id=self.order_id,
            openapi_service=self.__openapi_service,
            name=self._name,
        )

        logger.debug("Order created: %s", order)
        logger.debug("Adding datapoints to the order.")

        if self.__dataset:
            with tracer.start_as_current_span("add_datapoints"):
                _, failed_uploads = self.__dataset.add_datapoints(self.__datapoints)

                if failed_uploads:
                    raise FailedUploadException(self.__dataset, order, failed_uploads)

        else:
            raise RuntimeError(
                f"No dataset created for this order. order_id: {self.order_id}"
            )

        logger.debug("Datapoints added to the order.")
        logger.debug("Setting order to preview")
        try:
            self.__openapi_service.order_api.order_order_id_preview_post(self.order_id)
        except Exception:
            failed_uploads = _parse_failed_uploads(
                self.__openapi_service.dataset_api.dataset_dataset_id_datapoints_failed_get(
                    self.__dataset.id
                )
            )
            logger.error(
                "Internal download error for datapoints: %s\nWARNING: Failed Datapoints in error do not contain metadata.",
                failed_uploads,
            )
            raise FailedUploadException(self.__dataset, order, failed_uploads)
        return order

    def _workflow(self, workflow: Workflow) -> "RapidataOrderBuilder":
        """
        Set the workflow for the order.

        Args:
            workflow (Workflow): The workflow to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(workflow, Workflow):
            raise TypeError("Workflow must be of type Workflow.")

        self.__workflow = workflow
        return self

    def _referee(self, referee: Referee) -> "RapidataOrderBuilder":
        """
        Set the referee for the order.

        Args:
            referee (Referee): The referee to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(referee, Referee):
            raise TypeError("Referee must be of type Referee.")

        self.__referee = referee
        return self

    def _datapoints(
        self,
        datapoints: list[Datapoint],
    ) -> "RapidataOrderBuilder":
        """
        Set the datapoints for the order.

        Args:
            datapoints: (Sequence[Datapoint]): The datapoints to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(datapoints, list):
            raise TypeError(
                "Datapoints must be provided as a list of Datapoint objects."
            )

        self.__datapoints = datapoints
        return self

    def _settings(self, settings: Sequence[RapidataSetting]) -> "RapidataOrderBuilder":
        """
        Set the settings for the order.

        Args:
            settings (Settings): The settings to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """

        if not isinstance(settings, list):
            raise TypeError("Settings must be provided as a list of Setting objects.")

        for s in settings:
            if not isinstance(s, RapidataSetting):
                raise TypeError("The settings list must only contain Setting objects.")

        self.__settings = settings
        return self

    def _filters(self, filters: Sequence[RapidataFilter]) -> "RapidataOrderBuilder":
        """
        Set the filters for the order, e.g., country, language, userscore, etc.

        Args:
            filters (Sequence[Filters]): The user filters to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(filters, list):
            raise TypeError("Filters must be provided as a list of Filter objects.")

        for f in filters:
            if not isinstance(f, RapidataFilter):
                raise TypeError("Filters must be of type Filter.")

        if len(self.__user_filters) > 0:
            managed_print("Overwriting existing user filters.")

        self.__user_filters = filters
        return self

    def _validation_set_id(
        self, validation_set_id: str | None = None
    ) -> "RapidataOrderBuilder":
        """
        Set the validation set ID for the order.

        Args:
            validation_set_id (str): The validation set ID to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if validation_set_id is not None and not isinstance(validation_set_id, str):
            raise TypeError("Validation set ID must be of type str.")

        self.__validation_set_id = validation_set_id
        return self

    def _rapids_per_bag(self, amount: int) -> "RapidataOrderBuilder":
        """
        Define the number of tasks a user sees in a single session.

        Args:
            amount (int): The number of tasks a user sees in a single session.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.

        Raises:
            NotImplementedError: This method is not implemented yet.
        """
        raise NotImplementedError("Not implemented yet.")

    def _selections(
        self, selections: Sequence[RapidataSelection]
    ) -> "RapidataOrderBuilder":
        """
        Set the selections for the order.

        Args:
            selections (Sequence[Selection]): The selections to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(selections, list):
            raise TypeError(
                "Selections must be provided as a list of Selection objects."
            )

        for selection in selections:
            if not isinstance(selection, RapidataSelection):
                raise TypeError("Selections must be of type Selection.")

        self.__selections = selections  # type: ignore
        return self

    def _priority(self, priority: int | None = None) -> "RapidataOrderBuilder":
        """
        Set the priority for the order.

        Args:
            priority (int): The priority to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if priority is not None and not isinstance(priority, int):
            raise TypeError("Priority must be of type int.")

        self.__priority = priority
        return self

    def _sticky_state(
        self, sticky_state: Literal["None", "Temporary", "Permanent"] | None = None
    ) -> "RapidataOrderBuilder":
        """
        Set the sticky state for the order.
        """
        if sticky_state is not None and sticky_state not in [
            "None",
            "Temporary",
            "Permanent",
        ]:
            raise TypeError(
                "Sticky state must be of type Literal['None', 'Temporary', 'Permanent']."
            )

        self.__sticky_state_value = sticky_state
        return self



from typing import Sequence, cast
from rapidata.rapidata_client.datapoints.assets import (
    MediaAsset,
    TextAsset,
    MultiAsset,
    BaseAsset,
)
from rapidata.rapidata_client.datapoints.assets.constants import (
    ALLOWED_VIDEO_EXTENSIONS,
    ALLOWED_IMAGE_EXTENSIONS,
    ALLOWED_AUDIO_EXTENSIONS,
)
from rapidata.rapidata_client.datapoints.metadata import Metadata
from rapidata.api_client.models.dataset_dataset_id_datapoints_post_request_metadata_inner import (
    DatasetDatasetIdDatapointsPostRequestMetadataInner,
)
from rapidata.api_client.models.create_datapoint_from_text_sources_model import (
    CreateDatapointFromTextSourcesModel,
)
from pydantic import StrictStr, StrictBytes
from rapidata.api_client.models.asset_type import AssetType
from rapidata.api_client.models.prompt_type import PromptType
from rapidata.rapidata_client.datapoints.metadata._media_asset_metadata import (
    MediaAssetMetadata,
)
from rapidata.rapidata_client.datapoints.metadata._prompt_metadata import PromptMetadata
from rapidata.rapidata_client.config import logger


class Datapoint:
    def __init__(
        self,
        asset: MediaAsset | TextAsset | MultiAsset,
        metadata: Sequence[Metadata] | None = None,
    ):
        if not isinstance(asset, (MediaAsset, TextAsset, MultiAsset)):
            raise TypeError(
                "Asset must be of type MediaAsset, TextAsset, or MultiAsset."
            )

        if metadata and not isinstance(metadata, Sequence):
            raise TypeError("Metadata must be a list of Metadata objects.")

        if metadata and not all(isinstance(m, Metadata) for m in metadata):
            raise TypeError("All metadata objects must be of type Metadata.")

        self.asset = asset
        self.metadata = metadata

    def _get_effective_asset_type(self) -> type:
        """Get the effective asset type, handling MultiAsset by looking at its first asset."""
        if isinstance(self.asset, MultiAsset):
            return type(self.asset.assets[0])
        return type(self.asset)

    def is_media_asset(self) -> bool:
        """Check if this datapoint contains media assets."""
        effective_type = self._get_effective_asset_type()
        return issubclass(effective_type, MediaAsset)

    def is_text_asset(self) -> bool:
        """Check if this datapoint contains text assets."""
        effective_type = self._get_effective_asset_type()
        return issubclass(effective_type, TextAsset)

    def get_asset_type(self) -> AssetType:
        """Get the asset type of the datapoint."""
        if self.is_text_asset():
            return AssetType.TEXT
        elif self.is_media_asset():
            if isinstance(self.asset, MultiAsset):
                asset = self.asset.assets[0]
            else:
                asset = self.asset
            assert isinstance(asset, MediaAsset)
            if any(asset.path.endswith(ext) for ext in ALLOWED_IMAGE_EXTENSIONS):
                return AssetType.IMAGE
            elif any(asset.path.endswith(ext) for ext in ALLOWED_VIDEO_EXTENSIONS):
                return AssetType.VIDEO
            elif any(asset.path.endswith(ext) for ext in ALLOWED_AUDIO_EXTENSIONS):
                return AssetType.AUDIO
            else:
                logger.debug(
                    f"Cannot get asset type for asset type: {type(self.asset)}"
                )
                return AssetType.NONE
        else:
            logger.debug(f"Cannot get asset type for asset type: {type(self.asset)}")
            return AssetType.NONE

    def get_prompt_type(self) -> list[PromptType]:
        """Get the prompt type of the datapoint."""
        prompt_types = []
        for metadata in self.metadata or []:
            if isinstance(metadata, MediaAssetMetadata):
                prompt_types.append(PromptType.ASSET)
            elif isinstance(metadata, PromptMetadata):
                prompt_types.append(PromptType.TEXT)

        if len(prompt_types) == 0:
            return [PromptType.NONE]

        return prompt_types

    def get_texts(self) -> list[str]:
        """Extract text content from the asset(s)."""
        if isinstance(self.asset, TextAsset):
            return [self.asset.text]
        elif isinstance(self.asset, MultiAsset):
            texts = []
            for asset in self.asset.assets:
                if isinstance(asset, TextAsset):
                    texts.append(asset.text)
            return texts
        else:
            raise ValueError(f"Cannot extract text from asset type: {type(self.asset)}")

    def get_media_assets(self) -> list[MediaAsset]:
        """Extract media assets from the datapoint."""
        if isinstance(self.asset, MediaAsset):
            return [self.asset]
        elif isinstance(self.asset, MultiAsset):
            media_assets = []
            for asset in self.asset.assets:
                if isinstance(asset, MediaAsset):
                    media_assets.append(asset)
            return media_assets
        else:
            raise ValueError(
                f"Cannot extract media assets from asset type: {type(self.asset)}"
            )

    def get_local_file_paths(
        self,
    ) -> list[StrictStr | tuple[StrictStr, StrictBytes] | StrictBytes]:
        """Get local file paths for media assets that are stored locally."""
        if not self.is_media_asset():
            return []

        media_assets = self.get_media_assets()
        return [asset.to_file() for asset in media_assets if asset.is_local()]

    def get_urls(self) -> list[str]:
        """Get URLs for media assets that are remote."""
        if not self.is_media_asset():
            return []

        media_assets = self.get_media_assets()
        return [asset.path for asset in media_assets if not asset.is_local()]

    def get_prepared_metadata(
        self,
    ) -> list[DatasetDatasetIdDatapointsPostRequestMetadataInner]:
        """Prepare metadata for API upload."""
        metadata: list[DatasetDatasetIdDatapointsPostRequestMetadataInner] = []
        if self.metadata:
            for meta in self.metadata:
                meta_model = meta.to_model() if meta else None
                if meta_model:
                    metadata.append(
                        DatasetDatasetIdDatapointsPostRequestMetadataInner(meta_model)
                    )
        return metadata

    def create_text_upload_model(
        self, index: int
    ) -> CreateDatapointFromTextSourcesModel:
        """Create the model for uploading text datapoints."""
        if not self.is_text_asset():
            raise ValueError("Cannot create text upload model for non-text asset")

        texts = self.get_texts()
        metadata = self.get_prepared_metadata()

        return CreateDatapointFromTextSourcesModel(
            textSources=texts,
            sortIndex=index,
            metadata=metadata,
        )

    def get_datapoint_string(self) -> str:
        """Get the datapoint string for the datapoint."""
        if isinstance(self.asset, MediaAsset):
            return self.asset.path
        elif isinstance(self.asset, TextAsset):
            return self.asset.text
        else:
            raise ValueError(
                f"Cannot get datapoint string for asset type: {type(self.asset)}"
            )

    def __str__(self):
        return f"Datapoint(asset={self.asset})"

    def __repr__(self):
        return self.__str__()



import threading
import time
from tqdm import tqdm

from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.config import logger, rapidata_config


class ProgressTracker:
    """
    Track dataset upload progress in a background thread with shallow indentation.

    This class encapsulates the progress polling loop to keep methods in
    `RapidataDataset` simpler and below the maximum indentation depth.
    """

    def __init__(
        self,
        dataset_id: str,
        openapi_service: OpenAPIService,
        total_uploads: int,
        progress_poll_interval: float,
    ) -> None:
        self.dataset_id = dataset_id
        self.openapi_service = openapi_service
        self.total_uploads = total_uploads
        self.progress_poll_interval = progress_poll_interval
        self.upload_complete = False

    def _get_progress_or_none(self):
        try:
            return self.openapi_service.dataset_api.dataset_dataset_id_progress_get(
                self.dataset_id
            )
        except Exception:
            return None

    def complete(self) -> None:
        logger.debug("Upload complete, setting upload_complete to True")
        self.upload_complete = True

    def run(self) -> None:
        try:
            with tqdm(
                total=self.total_uploads,
                desc="Uploading datapoints",
                disable=rapidata_config.logging.silent_mode,
            ) as pbar:
                final_pass = False
                while True:
                    current_progress = self._get_progress_or_none()
                    if current_progress is None:
                        time.sleep(self.progress_poll_interval)
                        logger.debug(
                            "No progress yet, sleeping for %s seconds",
                            self.progress_poll_interval,
                        )
                        continue

                    total_completed = current_progress.ready + current_progress.failed

                    pbar.n = total_completed
                    pbar.refresh()

                    time.sleep(self.progress_poll_interval)
                    if total_completed >= self.total_uploads:
                        break

                    if self.upload_complete and current_progress.pending == 0:
                        if not final_pass:
                            logger.debug("Final pass")
                            time.sleep(self.progress_poll_interval)
                            final_pass = True
                            continue
                        logger.debug("Final pass done, breaking out of loop")
                        break

                pbar.close()

                success_rate = (
                    round((current_progress.ready / self.total_uploads * 100), 2)
                    if self.total_uploads > 0
                    else 0
                )

                logger.info(
                    "Upload complete: %s ready, %s failed, %s pending (%s%% success rate)",
                    current_progress.ready,
                    current_progress.failed,
                    current_progress.pending,
                    success_rate,
                )
        except Exception as e:
            logger.error("Progress tracking thread error: %s", str(e))
            raise RuntimeError("Progress tracking failed, aborting uploads")

    def create_thread(self) -> threading.Thread:
        thread = threading.Thread(target=self.run)
        thread.daemon = True
        return thread



from typing import Any
from rapidata.rapidata_client.filter._base_filter import RapidataFilter
from rapidata.api_client.models.or_user_filter_model import OrUserFilterModel
from rapidata.api_client.models.and_user_filter_model_filters_inner import (
    AndUserFilterModelFiltersInner,
)


class OrFilter(RapidataFilter):
    """A filter that combines multiple filters with a logical OR operation.
    This class implements a logical OR operation on a list of filters, where the condition is met if any of the filters' conditions are met.

    Args:
        filters (list[RapidataFilter]): A list of filters to be combined with OR.

    Example:
        ```python
        from rapidata import OrFilter, LanguageFilter, CountryFilter

        OrFilter([LanguageFilter(["en"]), CountryFilter(["US"])])
        ```

        This will match users who either have their phone set to English OR are located in the United States.
    """

    def __init__(self, filters: list[RapidataFilter]):
        if not all(isinstance(filter, RapidataFilter) for filter in filters):
            raise ValueError("Filters must be a RapidataFilter object")

        self.filters = filters

    def _to_model(self):
        return OrUserFilterModel(
            _t="OrFilter",
            filters=[
                AndUserFilterModelFiltersInner(filter._to_model())
                for filter in self.filters
            ],
        )

    def __str__(self) -> str:
        return f"OrFilter(filters={self.filters})"

    def __repr__(self) -> str:
        return f"OrFilter(filters={self.filters!r})"



from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.rapidata_client.datapoints.assets import TextAsset, MediaAsset
from rapidata.service import LocalFileService
from rapidata.service.openapi_service import OpenAPIService
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm

from typing import Generator
from rapidata.rapidata_client.config import logger
import time
import threading
from rapidata.rapidata_client.api.rapidata_api_client import (
    suppress_rapidata_error_logging,
)
from rapidata.rapidata_client.config.rapidata_config import rapidata_config
from rapidata.rapidata_client.order.dataset._progress_tracker import ProgressTracker

# Add OpenTelemetry context imports for thread propagation
from opentelemetry import context as otel_context


def chunk_list(lst: list, chunk_size: int) -> Generator:
    for i in range(0, len(lst), chunk_size):
        yield lst[i : i + chunk_size]


class RapidataDataset:
    def __init__(self, dataset_id: str, openapi_service: OpenAPIService):
        self.id = dataset_id
        self.openapi_service = openapi_service
        self.local_file_service = LocalFileService()

    def add_datapoints(
        self,
        datapoints: list[Datapoint],
    ) -> tuple[list[Datapoint], list[Datapoint]]:
        if not datapoints:
            return [], []

        effective_asset_type = datapoints[0]._get_effective_asset_type()

        logger.debug(f"Config for datapoint upload: {rapidata_config}")

        if issubclass(effective_asset_type, MediaAsset):
            return self._add_media_from_paths(
                datapoints,
            )
        elif issubclass(effective_asset_type, TextAsset):
            return self._add_texts(datapoints)
        else:
            raise ValueError(f"Unsupported asset type: {effective_asset_type}")

    def _add_texts(
        self, datapoints: list[Datapoint]
    ) -> tuple[list[Datapoint], list[Datapoint]]:

        def upload_text_datapoint(datapoint: Datapoint, index: int) -> Datapoint:
            model = datapoint.create_text_upload_model(index)

            self.openapi_service.dataset_api.dataset_dataset_id_datapoints_texts_post(
                dataset_id=self.id, create_datapoint_from_text_sources_model=model
            )
            return datapoint

        def upload_with_context(
            context: otel_context.Context, datapoint: Datapoint, index: int
        ) -> Datapoint:
            """Wrapper function that runs upload_text_datapoint with the provided context."""
            token = otel_context.attach(context)
            try:
                return upload_text_datapoint(datapoint, index)
            finally:
                otel_context.detach(token)

        successful_uploads: list[Datapoint] = []
        failed_uploads: list[Datapoint] = []

        # Capture the current OpenTelemetry context before creating threads
        current_context = otel_context.get_current()

        total_uploads = len(datapoints)
        with ThreadPoolExecutor(
            max_workers=rapidata_config.upload.maxWorkers
        ) as executor:
            future_to_datapoint = {
                executor.submit(
                    upload_with_context, current_context, datapoint, i
                ): datapoint
                for i, datapoint in enumerate(datapoints)
            }

            with tqdm(
                total=total_uploads,
                desc="Uploading text datapoints",
                disable=rapidata_config.logging.silent_mode,
            ) as pbar:
                for future in as_completed(future_to_datapoint.keys()):
                    datapoint = future_to_datapoint[future]
                    try:
                        result = future.result()
                        pbar.update(1)
                        successful_uploads.append(result)
                    except Exception as e:
                        failed_uploads.append(datapoint)
                        logger.error("Upload failed for %s: %s", datapoint, str(e))

        return successful_uploads, failed_uploads

    def _process_single_upload(
        self,
        datapoint: Datapoint,
        index: int,
    ) -> tuple[list[Datapoint], list[Datapoint]]:
        """
        Process single upload with retry logic and error tracking.

        Args:
            media_asset: MediaAsset or MultiAsset to upload
            meta_list: Optional sequence of metadata for the asset
            index: Sort index for the upload
            max_retries: Maximum number of retry attempts (default: 3)

        Returns:
            tuple[list[Datapoint], list[Datapoint]]: Lists of successful and failed datapoints
        """
        logger.debug("Processing single upload for %s with index %s", datapoint, index)

        local_successful: list[Datapoint] = []
        local_failed: list[Datapoint] = []

        metadata = datapoint.get_prepared_metadata()

        local_paths = datapoint.get_local_file_paths()
        urls = datapoint.get_urls()

        last_exception = None
        for attempt in range(rapidata_config.upload.maxRetries):
            try:
                with suppress_rapidata_error_logging():
                    self.openapi_service.dataset_api.dataset_dataset_id_datapoints_post(
                        dataset_id=self.id,
                        file=local_paths,
                        url=urls,
                        metadata=metadata,
                        sort_index=index,
                    )

                local_successful.append(datapoint)

                return local_successful, local_failed

            except Exception as e:
                last_exception = e
                if attempt < rapidata_config.upload.maxRetries - 1:
                    # Exponential backoff: wait 1s, then 2s, then 4s
                    retry_delay = 2**attempt
                    time.sleep(retry_delay)
                    logger.debug("Error: %s", str(last_exception))
                    logger.debug(
                        "Retrying %s of %s...",
                        attempt + 1,
                        rapidata_config.upload.maxRetries,
                    )

        # If we get here, all retries failed
        local_failed.append(datapoint)
        tqdm.write(
            f"Upload failed for {datapoint} after {rapidata_config.upload.maxRetries} attempts. \nFinal error: \n{str(last_exception)}"
        )

        return local_successful, local_failed

    def _process_uploads_in_chunks(
        self,
        datapoints: list[Datapoint],
    ) -> tuple[list[Datapoint], list[Datapoint]]:
        """
        Process uploads in chunks with a ThreadPoolExecutor.

        Args:
            media_paths: List of assets to upload
            multi_metadata: Optional sequence of sequences of metadata
            chunk_size: Number of items to process in each batch

        Returns:
            tuple[list[str], list[str]]: Lists of successful and failed uploads
        """
        successful_uploads: list[Datapoint] = []
        failed_uploads: list[Datapoint] = []

        def process_upload_with_context(
            context: otel_context.Context, datapoint: Datapoint, index: int
        ) -> tuple[list[Datapoint], list[Datapoint]]:
            """Wrapper function that runs _process_single_upload with the provided context."""
            token = otel_context.attach(context)
            try:
                return self._process_single_upload(datapoint, index)
            finally:
                otel_context.detach(token)

        # Capture the current OpenTelemetry context before creating threads
        current_context = otel_context.get_current()

        with ThreadPoolExecutor(
            max_workers=rapidata_config.upload.maxWorkers
        ) as executor:
            # Process uploads in chunks to avoid overwhelming the system
            for chunk_idx, chunk in enumerate(
                chunk_list(datapoints, rapidata_config.upload.chunkSize)
            ):
                futures = [
                    executor.submit(
                        process_upload_with_context,
                        current_context,
                        datapoint,
                        chunk_idx * rapidata_config.upload.chunkSize + i,
                    )
                    for i, datapoint in enumerate(chunk)
                ]

                # Wait for this chunk to complete before starting the next one
                for future in as_completed(futures):
                    try:
                        chunk_successful, chunk_failed = future.result()
                        successful_uploads.extend(chunk_successful)
                        failed_uploads.extend(chunk_failed)
                    except Exception as e:
                        logger.error("Future execution failed: %s", str(e))

        return successful_uploads, failed_uploads

    def _add_media_from_paths(
        self,
        datapoints: list[Datapoint],
        progress_poll_interval: float = 0.5,
    ) -> tuple[list[Datapoint], list[Datapoint]]:
        """
        Upload media paths in chunks with managed resources.

        Args:
            datapoints: List of Datapoint objects to upload
            chunk_size: Number of items to process in each batch
            progress_poll_interval: Time in seconds between progress checks
        Returns:
            tuple[list[Datapoint], list[Datapoint]]: Lists of successful and failed datapoints

        Raises:
            ValueError: If multi_metadata lengths don't match media_paths length
        """

        # Setup tracking variables
        total_uploads = len(datapoints)

        # Create and start progress tracking thread
        progress_tracker = ProgressTracker(
            dataset_id=self.id,
            openapi_service=self.openapi_service,
            total_uploads=total_uploads,
            progress_poll_interval=progress_poll_interval,
        )
        progress_thread = progress_tracker.create_thread()
        progress_thread.start()

        # Process uploads in chunks
        try:
            successful_uploads, failed_uploads = self._process_uploads_in_chunks(
                datapoints,
            )
        finally:
            progress_tracker.complete()
            progress_thread.join(10)

        if failed_uploads:
            logger.error(
                "Upload failed for %s datapoints: %s",
                len(failed_uploads),
                failed_uploads,
            )

        return successful_uploads, failed_uploads

    def __str__(self) -> str:
        return f"RapidataDataset(id={self.id})"

    def __repr__(self) -> str:
        return self.__str__()



from rapidata.api_client import (
    CompareWorkflowModelPairMakerConfig,
    OnlinePairMakerConfigModel,
    EloConfigModel,
)
from rapidata.api_client.models.compare_workflow_model import CompareWorkflowModel
from rapidata.rapidata_client.workflow._base_workflow import Workflow
from rapidata.rapidata_client.datapoints.metadata import PromptMetadata
from rapidata.api_client.models.dataset_dataset_id_datapoints_post_request_metadata_inner import (
    DatasetDatasetIdDatapointsPostRequestMetadataInner,
)
from rapidata.api_client import ComparePayload
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.api_client.models.rapid_modality import RapidModality


class RankingWorkflow(Workflow):
    modality = RapidModality.COMPARE

    def __init__(
        self,
        criteria: str,
        total_comparison_budget: int,
        random_comparisons_ratio,
        elo_start: int = 1200,
        elo_k_factor: int = 40,
        elo_scaling_factor: int = 400,
        context: str | None = None,
    ):
        super().__init__(type="CompareWorkflowConfig")

        self.context = (
            [
                DatasetDatasetIdDatapointsPostRequestMetadataInner(
                    PromptMetadata(context).to_model()
                )
            ]
            if context
            else None
        )

        self.criteria = criteria
        self.total_comparison_budget = total_comparison_budget
        self.random_comparisons_ratio = random_comparisons_ratio
        self.elo_start = elo_start
        self.elo_k_factor = elo_k_factor
        self.elo_scaling_factor = elo_scaling_factor

        self.pair_maker_config = CompareWorkflowModelPairMakerConfig(
            OnlinePairMakerConfigModel(
                _t="OnlinePairMaker",
                totalComparisonBudget=total_comparison_budget,
                randomMatchesRatio=random_comparisons_ratio,
            )
        )

        self.elo_config = EloConfigModel(
            startingElo=elo_start,
            kFactor=elo_k_factor,
            scalingFactor=elo_scaling_factor,
        )

    def _to_model(self) -> CompareWorkflowModel:

        return CompareWorkflowModel(
            _t="CompareWorkflow",
            criteria=self.criteria,
            eloConfig=self.elo_config,
            pairMakerConfig=self.pair_maker_config,
            metadata=self.context,
        )

    def _to_payload(self, datapoint: Datapoint) -> ComparePayload:
        return ComparePayload(
            _t="ComparePayload",
            criteria=self.criteria,
        )

    def __str__(self) -> str:
        return f"RankingWorkflow(criteria='{self.criteria}', context={self.context})"

    def __repr__(self) -> str:
        return f"RankingWorkflow(criteria={self.criteria!r}, total_comparison_budget={self.total_comparison_budget!r}, random_comparisons_ratio={self.random_comparisons_ratio!r}, elo_start={self.elo_start!r}, elo_k_factor={self.elo_k_factor!r}, elo_scaling_factor={self.elo_scaling_factor!r}, context={self.context!r})"



from typing import Any
from rapidata.rapidata_client.filter._base_filter import RapidataFilter
from rapidata.api_client.models.and_user_filter_model import AndUserFilterModel
from rapidata.api_client.models.and_user_filter_model_filters_inner import (
    AndUserFilterModelFiltersInner,
)


class AndFilter(RapidataFilter):
    """A filter that combines multiple filters with a logical AND operation.
    This class implements a logical AND operation on a list of filters, where the condition is met if all of the filters' conditions are met.

    Args:
        filters (list[RapidataFilter]): A list of filters to be combined with AND.

    Example:
        ```python
        from rapidata import AndFilter, LanguageFilter, CountryFilter

        AndFilter([LanguageFilter(["en"]), CountryFilter(["US"])])
        ```

        This will match users who have their phone set to English AND are located in the United States.
    """

    def __init__(self, filters: list[RapidataFilter]):
        if not all(isinstance(filter, RapidataFilter) for filter in filters):
            raise ValueError("Filters must be a RapidataFilter object")

        self.filters = filters

    def _to_model(self):
        return AndUserFilterModel(
            _t="AndFilter",
            filters=[
                AndUserFilterModelFiltersInner(filter._to_model())
                for filter in self.filters
            ],
        )

    def __str__(self) -> str:
        return f"AndFilter(filters={self.filters})"

    def __repr__(self) -> str:
        return f"AndFilter(filters={self.filters!r})"



from typing import cast
from rapidata.api_client.models.file_asset_model import FileAssetModel
from rapidata.api_client.models.get_failed_datapoints_result import (
    GetFailedDatapointsResult,
)
from rapidata.api_client.models.multi_asset_model import MultiAssetModel
from rapidata.api_client.models.original_filename_metadata_model import (
    OriginalFilenameMetadataModel,
)
from rapidata.api_client.models.source_url_metadata_model import SourceUrlMetadataModel
from rapidata.rapidata_client.datapoints.assets import MediaAsset, MultiAsset
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.rapidata_client.order.dataset._rapidata_dataset import RapidataDataset
from rapidata.rapidata_client.order.rapidata_order import RapidataOrder


class FailedUploadException(Exception):
    """Custom error class for Failed Uploads to the Rapidata order."""

    def __init__(
        self,
        dataset: RapidataDataset,
        order: RapidataOrder,
        failed_uploads: list[Datapoint],
    ):
        self.dataset = dataset
        self.order = order
        self.failed_uploads = failed_uploads

    def __str__(self) -> str:
        return f"Failed to upload {self.failed_uploads}"


def _parse_failed_uploads(failed_uploads: GetFailedDatapointsResult) -> list[Datapoint]:
    failed_datapoints = failed_uploads.datapoints
    if not failed_datapoints:
        return []
    if isinstance(failed_datapoints[0].asset.actual_instance, FileAssetModel):
        failed_assets = [
            MediaAsset(
                __get_asset_name(cast(FileAssetModel, datapoint.asset.actual_instance))
            )
            for datapoint in failed_datapoints
        ]
    elif isinstance(failed_datapoints[0].asset.actual_instance, MultiAssetModel):
        failed_assets = []
        backend_assets = [
            cast(MultiAssetModel, failed_upload.asset.actual_instance).assets
            for failed_upload in failed_datapoints
        ]
        for assets in backend_assets:
            failed_assets.append(
                MultiAsset(
                    [
                        MediaAsset(
                            __get_asset_name(
                                cast(FileAssetModel, asset.actual_instance)
                            )
                        )
                        for asset in assets
                        if isinstance(asset.actual_instance, FileAssetModel)
                    ]
                )
            )
    else:
        raise ValueError(
            f"Unsupported asset type: {type(failed_datapoints[0].asset.actual_instance)}"
        )

    return [Datapoint(asset=asset) for asset in failed_assets]


def __get_asset_name(failed_datapoint: FileAssetModel) -> str:
    metadata = failed_datapoint.metadata
    if "sourceUrl" in metadata:
        return cast(SourceUrlMetadataModel, metadata["sourceUrl"].actual_instance).url
    elif "originalFilename" in metadata:
        return cast(
            OriginalFilenameMetadataModel, metadata["originalFilename"].actual_instance
        ).original_filename
    else:
        return ""
